"""wsgimagic is designed to allow you to effortlessly transition any WSGI compliant Python application
(eg Flask, Django) to a completely serverless architecture using an AWS APIGateway Lambda Proxy pointing to a Lambda
function running your code. By simply using the wsgi_magic decorator, you can pass the incoming request off to your
application and ensure that the values are returned in the required format.
"""

import sys
from io import StringIO
from datetime import datetime
from functools import wraps
import base64
from magic_core import TranslatedRequest, RawResponse, WSGIHandler


def _map_api_gateway_to_request(resource: str, path: str, httpMethod: str, headers: dict,
                                multiValueHeaders: dict, queryStringParameters: dict,
                                multiValueQueryStringParameters: dict, pathParameters: dict,
                                stageVariables: dict, requestContext: dict, body: str,
                                isBase64Encoded: bool, **kwargs):
    """Maps the incoming event from API Gateway to the necessary request structure for our
    application. Note the use of the addition **kwargs in case Amazon adds new fields.
    """
    mapped_headers = {'HTTP_'+key.upper(): value for key, value in headers.items()}
    if multiValueHeaders is not None:
        mapped_headers.update({'HTTP_'+key.upper(): value[0] for key, value in
                               multiValueHeaders.items()})
    request_body = body
    if isBase64Encoded:
        request_body = base64.b64decode(request_body)

    query_string = None
    if len(queryStringParameters) > 0:
        query_string = '&'.join(['{0}={1}'.format(key, value) for key, value
                                 in queryStringParameters.items()])
    if multiValueQueryStringParameters is not None:
        additional_query_string = '&'.join(['{0}={1}'.format(key, value[0]) for key, value
                                 in multiValueQueryStringParameters.items()])
        if query_string is None:
            query_string = additional_query_string
        else:
            query_string += '&' + additional_query_string

    return TranslatedRequest(path, httpMethod, mapped_headers, query_string, request_body)


def _basic_error_handler(exception: Exception) -> dict:
    """This is a basic error handler that just says something went wrong. Make sure your handlers
    have the same signature!
    """
    return {'statusCode': '500',
            'headers': {'Date': datetime.now().strftime("%a, %d %b %Y %H:%M:%S EST"),
                        'Server': 'WSGIMagic'},
            'body': 'Server Error'}


def _build_proxy_response(response: RawResponse, error_handler: callable) -> dict:
    """Once the application completes the request, maps the results into the format required by
    AWS.
    """
    try:
        if response.caught_exception is not None:
            raise response.caught_exception
        message = ''.join([str(message) for message in response.result])

    except Exception as e:
        return error_handler(e)
    return {'statusCode': response.response_status.split(' ')[0],
            'headers': response.outbound_headers,
            'body': message}


def lambda_magic(wsgi_application, additional_response_headers: dict=dict(), server: str='',
                 port: int=0, error_handler=_basic_error_handler, response_modifier: callable=None):
    """This is the magical decorator that handles all of your Lambda WSGI application needs!

    Keyword Args:
        wsgi_application: The application that will be fed the wsgi request.
        additional_headers: This is used to pass along any addition headers that you may need to
                            send to the client
        server: The server host name. This is only important if you are using it in your app.
        port: Since we aren't actually going to be binding to a port, this is mildly falsified, but
              us it if you need it!
        error_handler: This is a callable that is used to return server error messages if something
                       goes really wrong. If you are implementing your own, make absolutely sure
                       that your function signature matches that of _basic_error_handler, otherwise
                       you'll fail to send an error, which is very embarrassing.
        response_modifier: This is an optional callable that you can enter to act on the response
                           from WSGIHandler. Make sure that this function is returning an instance
                           of RawResponse, otherwise the _build_proxy_response function will error.
        """
    def internal(lambda_handler):
        @wraps(lambda_handler)
        def handle(*arg, **kwargs):
            lambda_handler(*arg, **kwargs)
            formatted_request = _map_api_gateway_to_request(**arg[0])
            requester = WSGIHandler(wsgi_application, additional_response_headers, server, port)
            response = requester.handle_request(formatted_request)
            if response_modifier is not None:
                response = response_modifier(response)
            return _build_proxy_response(response, error_handler)
        return handle
    return internal



import sys
from io import StringIO
from datetime import datetime as dt


class TranslatedRequest:
    """All sub modules should translate their incoming request object into this form to be passed
    off to the WSGIHandler.
    """
    def __init__(self, path: str, http_method: str, headers: dict, query_string: str,
                 body: str):
        self.path = path
        self.http_method = http_method
        self.headers = headers
        self.query_string= query_string
        self.body = body


class RawResponse:
    """All submodules should expect to receive an instance of this class back from the
    handle_response method of the WSGIHandler. all outgoing response mappers can then be designed to
    build from this.
    """
    def __init__(self):
        self.result: 'iterable' = None
        self.caught_exception: Exception = None
        self.response_status: str = None
        self.outbound_headers: dict = dict()


class WSGIHandler:
    """This class performs the heavy lifting of translating incoming requests and returning
    responses.
    """

    def __init__(self, wsgi_application: 'WSGI Application', additional_response_headers: dict,
                 server: str, port: int):
        """ Initializer for the WSGIHandler class.

        Keyword Args:
        wsgi_application: The application that will be fed the wsgi request.
        additional_headers: This is used to pass along any addition headers that you may need to
                            send to the client
        server: The server host name. This is only important if you are using it in your app.
        port: Since we aren't actually going to be binding to a port, this is mildly falsified, but
              use it if you need it!
        """

        self.app = wsgi_application
        self.additional_response_headers = additional_response_headers
        self.server = server
        self.port = port
        self.response = RawResponse()

    @staticmethod
    def generate_env_dict(request: TranslatedRequest, server: str, port: int) -> dict:
        """Builds the necessary WSGI environment based on the incoming request"""
        environment = dict()

        environment['wsgi.version'] = (1, 0)
        environment['wsgi.url_scheme'] = 'http'
        environment['wsgi.input'] = StringIO(request.body)
        environment['wsgi.errors'] = sys.stderr
        environment['wsgi.multithread'] = False
        environment['wsgi.multiprocess'] = False
        environment['wsgi.run_once'] = False
        environment['REQUEST_METHOD'] = request.http_method
        environment['PATH_INFO'] = request.path
        environment['SERVER_NAME'] = server
        environment['SERVER_PORT'] = str(port)
        if request.query_string is not None:
            environment['QUERY_STRING'] = request.query_string
        else:
            environment['QUERY_STRING'] = ''
        environment.update(request.headers)
        return environment

    def wsgi_callback(self, status: str, response_headers: [()], exc_info=None) -> None:
        """This is used as the start response function that is sent to the application."""
        try:
            if exc_info is not None:
                raise(exc_info[0], exc_info[1], exc_info[2])
            response_header_dict = {tup[0]: tup[1] for tup in response_headers}
            if self.additional_response_headers is not None:
                response_header_dict.update(self.additional_response_headers)
            response_header_dict.update({'Date': dt.now().strftime("%a, %d %b %Y %H:%M:%S EST"),
                                         'Server': 'WSGIMagic'})
            self.response.response_status = status
            self.response.outbound_headers = response_header_dict

        except Exception as e:
            self.response.caught_exception = e

    def handle_request(self, request: TranslatedRequest) -> RawResponse:
        self.response.result = self.app(self.generate_env_dict(request, self.server, self.port),
                                        self.wsgi_callback)
        return self.response



from setuptools import setup, find_packages

with open('README.md', 'r') as readme:
    long_desc = readme.read()

setup(name='wsgimagic',
      version='1.0.0',
      description='Serverless WSGI apps made easy',
      packages=find_packages(exclude=('tests',)),
      author="Kyle Hinton",
      license="MIT",
      long_description=long_desc,
      long_description_content_type='text/markdown')

